"""Implementation of CLI check command."""

import os
import json
import warnings

import numpy as np

from treem.io import SWC, TreemEncoder


def check(args):
    """Checks morphology reconstruction for structural consistency."""
    # pylint: disable=too-many-branches
    # pylint: disable=too-many-statements
    data = None
    err = dict()
    while True:
        if not os.path.exists(args.file) or not os.path.isfile(args.file):
            err['no_file'] = [args.file]
            break
        if not args.file.lower().endswith('swc'):
            err['not_swc_ext'] = [args.file.split('.')[-1]]
        try:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                data = np.loadtxt(args.file)
        except ValueError:
            err['not_array'] = [True]
            break
        if data.shape[0] == 0:
            err['no_data'] = [True]
            break
        if len(data.shape) == 1:
            err['single_point'] = [data[SWC.I].astype(int)]
            break
        if data.shape[1] != len(SWC.COLS):
            err['not_swc_cols'] = [data.shape[1]]
            break
        first = data[0]
        if first[SWC.I] != 1:
            err['node1_not_id1'] = [first[SWC.I].astype(int)]
        if first[SWC.P] != -1:
            err['node1_has_parent'] = [first[SWC.P].astype(int)]
        if first[SWC.T] != SWC.SOMA:
            err['node1_not_soma'] = [first[SWC.T].astype(int)]
        types = set(data[:, SWC.T].astype(int))
        if not types.issubset(SWC.TYPES):
            err['not_valid_types'] =\
                [data[x][SWC.I].astype(int)
                 for t in types.difference(SWC.TYPES)
                 for x in np.where(data[:, SWC.T] == t)[0]]
        ids = data[:, SWC.I].astype(int)
        if not (ids > 0).all():
            err['not_valid_ids'] = ids[ids <= 0]
            break
        idp = data[1:, SWC.P].astype(int)
        if not (idp > 0).all():
            err['not_valid_parent_ids'] = ids[1:][idp <= 0]
            break
        ids_set = set(ids)
        idp_set = set(idp)
        if len(ids_set) != len(ids):
            seen = set()
            err['non_unique_ids'] =\
                {x for x in ids if x in seen or seen.add(x)}
            break
        if not idp_set.issubset(ids_set):
            err['undef_parent_ids'] =\
                [data[x][SWC.I].astype(int)
                 for p in idp_set.difference(ids_set)
                 for x in np.where(data[:, SWC.P] == p)[0]]
            break
        inc = ids[1:] - ids[:-1]
        if not (inc > 0).all():
            err['non_increasing_ids'] = ids[np.where(inc <= 0)[0] + 1]
            break
        if not (inc == 1).all():
            err['non_sequential_ids'] = ids[np.where(inc != 1)[0] + 1]
            break
        if not (ids[1:] > idp).all():
            err['non_descendant'] = ids[np.where(ids[1:] <= idp)[0] + 1]
            break
        if [data[x][SWC.I].astype(int) for t in types.difference([1])
                for x in np.where(data[:, SWC.T] == t)[0][:1]
                if data[x][SWC.P] != 1]:
            err['non_stem_neurite'] =\
                [data[x][SWC.I].astype(int)
                 for t in types.difference([1])
                 for x in np.where(data[:, SWC.T] == t)[0][:1]
                 if data[x][SWC.P] != 1]
        break

    if not args.quiet:
        for k in err:
            print(f'{k}:', end=' ')
            print(*err[k])

    if args.out:
        with open(args.out, 'w') as f:  # pylint: disable=invalid-name
            json.dump(err, f, cls=TreemEncoder)

    return len(err)



"""Basic tree data structure."""

from collections import deque


class Tree():
    """Recursive tree data structure."""

    def __init__(self):
        """Constructor of empty tree."""
        self.parent = None
        self.siblings = list()

    def add(self, tree):
        """Appends tree as continuation."""
        tree.parent = self
        self.siblings.append(tree)
        return tree

    def is_root(self):
        """Returns True if tree has no parent."""
        return self.parent is None

    def is_fork(self):
        """Returns True for a branching point (except root)."""
        return len(self.siblings) > 1 and not self.is_root()

    def is_leaf(self):
        """Returns True for a terminal node."""
        return len(self.siblings) == 0

    def preorder(self):
        """Traverses tree in pre-order (depth first).

        Yields:
            sequence of tree nodes (generator object).
        """
        queue = deque((self,))
        while queue:
            node = queue.pop()
            queue.extend(reversed(node.siblings))
            yield node

    def postorder(self):
        """Traverses tree in post-order (depth first).

        Yields:
            sequence of tree nodes (generator object).
        """
        queue = deque((self,))
        visited = set()
        while queue:
            node = queue[-1]
            if node not in visited:
                visited.add(node)
                queue.extend(reversed(node.siblings))
            else:
                queue.pop()
                yield node

    def levelorder(self):
        """Traverses tree in level-order (breadth first).

        Yields:
            sequence of tree nodes (generator object).
        """
        queue = deque((self,))
        visited = set()
        while queue:
            node = queue[-1]
            if node not in visited:
                visited.add(node)
                queue.extend(node.siblings)
            else:
                queue.pop()
                yield node

    def ascendorder(self):
        """Iterates tree nodes in ascending order until root is reached.

        Yields:
            sequence of tree nodes (generator object).
        """
        node = self
        while node is not None:
            yield node
            node = node.parent

    def leaves(self):
        """Iterates tree terminals.

        Returns:
            sequence of tree nodes (generator object).
        """
        return filter(Tree.is_leaf, self.preorder())

    def forks(self, iterator=preorder):
        """Iterates tree branching points.

        Args:
            iterator: iterator type (defaults to ``Tree.preorder``)

        Returns:
            sequence of tree nodes (generator object).
        """
        return filter(Tree.is_fork, iterator(self))

    def degree(self):
        """Returns node degree."""
        return len(self.siblings)

    def size(self):
        """Returns tree size."""
        return sum(1 for node in self.preorder())

    def breadth(self):
        """Returns tree breadth."""
        return sum(1 for node in self.leaves())

    def depth(self):
        """Returns node depth."""
        return sum(1 for node in self.ascendorder()) - 1

    level = depth

    def height(self):
        """Returns tree height."""
        path = list()
        node_depth = self.depth()
        for leaf in self.leaves():
            path.append(leaf.depth() - node_depth)
        return max(path)

    def width(self):
        """Returns tree width."""
        node_depth = self.depth()
        root = deque(self.ascendorder(), maxlen=1).pop()
        return sum(1 for node in root.preorder() if node.depth() == node_depth)



"""Implementation of CLI convert command."""

import numpy as np

from treem import Node, SWC, Morph


def convert(args):
    """Converts input data to compliant SWC format."""
    vprint = print if not args.quiet else lambda *a, **k: None
    nam = 'I', 'T', 'X', 'Y', 'Z', 'R', 'P'
    fmt = 'i', 'i', 'f', 'f', 'f', 'f', 'i'
    try:
        data = np.loadtxt(args.file, dtype={'names': nam, 'formats': fmt})
        types = args.type if args.type else SWC.TYPES
        nodes = [Node(row) for row in data if row[SWC.T] in types
                 or row[SWC.P] == -1]
        root = [node for node in nodes if node.v[SWC.P] == -1][0]
        root.v[SWC.T] = SWC.SOMA
        for i, node in enumerate(nodes, 1):
            parent_id = node.v[SWC.P]
            for parent in nodes:
                if parent.v[SWC.I] == parent_id:
                    parent.add(node)
                    break
            vprint(f'converted {int(i/len(nodes)*100):>3d}%', end='\r')
        vprint()
        idmap = {-1: -1}
        for ident, node in enumerate(root.walk(), 1):
            idmap[node.v[SWC.I]] = ident
        for row in data:
            if row[SWC.T] in types or row[SWC.P] == -1:
                row[SWC.I], row[SWC.P] = idmap[row[SWC.I]], idmap[row[SWC.P]]
        data = np.array([node.v for node in root.walk()])
        Morph(data=data).save(args.out)
        return 0
    except (KeyError, IndexError, ValueError):
        vprint(f'can not convert {args.file}.')
        return 1



"""Implementation of CLI repair command."""

import math

from itertools import chain

import numpy as np

from treem import Morph, SWC
from treem.utils.geom import repair_branch, sample, norm, rotation


def repair(args):
    """Corrects morphology reconstruction at the given nodes."""
    # pylint: disable=invalid-name
    # pylint: disable=too-many-locals
    # pylint: disable=too-many-branches
    # pylint: disable=too-many-statements
    # pylint: disable=cell-var-from-loop
    # pylint: disable=expression-not-assigned
    vprint = print if args.verbose else lambda *a, **k: None
    morph = Morph(args.file)
    err = 0

    if args.center:
        morph.data[:, SWC.XYZ] -= morph.root.coord()

    if args.translate:
        morph.data[:, SWC.XYZ] += np.array(args.translate)

    if args.rotate:
        morph.rotate([1, 0, 0], args.rotate[0]/180*math.pi)
        morph.rotate([0, 1, 0], args.rotate[1]/180*math.pi)
        morph.rotate([0, 0, 1], args.rotate[2]/180*math.pi)

    if args.shrink_xy:
        scale = np.array([args.shrink_xy, args.shrink_xy, 1])
        origin = morph.root.coord().copy()
        for stem in morph.stems():
            for node in stem.walk():
                coord = node.coord()
                coord *= scale
        shift = origin - morph.root.coord()
        morph.translate(shift)

    if args.shrink:
        if args.bottom_up:
            bottom = max(x.coord()[2] for x in morph.root.walk())
        else:
            bottom = min(x.coord()[2] for x in morph.root.walk())
        for node in morph.root.walk():
            z = node.coord()[2]
            z = bottom + args.shrink * (z - bottom)
            node.v[SWC.Z] = z

    if args.zjump:
        nodes = [x for x in morph.root.walk() if x.ident() in args.zjump]
        for node in nodes:
            jump = node.parent.coord()[2] - node.coord()[2]
            if args.zjump_mode == 'align':
                shift = [0, 0, jump]
                morph.translate(shift, node)
            elif args.zjump_mode == 'split':
                shift = [0, 0, jump/2]
                jump_sec = list(node.section())
                for split_node in jump_sec[:int(len(jump_sec)/2)]:
                    morph.move(shift, split_node)
            elif args.zjump_mode == 'tilt':
                parent = node.parent
                dist = max(norm(parent.coord()-jump_node.coord())
                           for jump_node in node.leaves())
                leng = max(norm(node.coord()-jump_node.coord())
                           for jump_node in node.leaves())
                leaf = [jump_node for jump_node in node.leaves()
                        if norm(node.coord()-jump_node.coord()) == leng][0]
                vdir = leaf.coord() - node.parent.coord()
                shift = [0, 0, jump*leng/dist]
                morph.translate(shift, node)
                udir = leaf.coord() - node.coord()
                axis, angle = rotation(udir, vdir)
                morph.rotate(axis, angle, node)
            elif args.zjump_mode == 'join':
                parent = node.parent
                dist = max(norm(parent.coord()-jump_node.coord())
                           for jump_node in node.leaves())
                leng = max(norm(node.coord()-jump_node.coord())
                           for jump_node in node.leaves())
                leaf = [jump_node for jump_node in node.leaves()
                        if norm(node.coord()-jump_node.coord()) == leng][0]
                vdir = leaf.coord() - node.parent.coord()
                shift = [0, 0, jump*leng/dist]
                morph.translate(shift, node)
                udir = leaf.coord() - node.coord()
                axis, angle = rotation(udir, vdir)
                morph.rotate(axis, angle, node)
                start = list(node.section(reverse=True))[-1].parent
                dist = max(norm(start.coord()-jump_node.coord())
                           for jump_node in node.leaves())
                leng = morph.length(node.parent.section(reverse=True))
                vdir = leaf.coord() - start.coord()
                shift = [0, 0, -jump*leng/dist]
                morph.translate(shift, node.parent)
                udir = leaf.coord() - node.coord()
                axis, angle = rotation(udir, vdir)
                morph.rotate(axis, angle, node)

    if args.pool:
        pool = [Morph(f) for f in args.pool]

    if args.diam:
        nodes = [x for x in morph.root.walk() if x.ident() in args.diam]
        if args.pool:
            types = {x.type() for x in nodes}
        if args.diam_mode == 'joint':
            for node in nodes:
                r = 0
                if node.parent.type() != SWC.SOMA:
                    r += node.parent.radius()
                    r /= 2
                if not node.is_fork():
                    r += node.siblings[0].radius()
                    r /= 2
                if r:
                    node.v[SWC.R] = r
                else:
                    vprint(f'diam in node {node.ident()} not repaired')
                    err += 1
        elif args.diam_mode == 'sec':
            for node in nodes:
                sec = list(node.section(reverse=True))
                sec = list(sec[-1].section())
                r = morph.radii(sec).mean()
                node.v[SWC.R] = r
        elif args.diam_mode == 'order':
            for node in nodes:
                point_type = node.type()
                order = node.order()
                if args.pool:
                    radii = [m.radii(sec).mean() for m in pool
                             for sec in m.root.sections()
                             if sec[0].type() in types
                             and sec[0].order() == order]
                    if radii:
                        r = np.mean(radii)
                        node.v[SWC.R] = r
                    else:
                        vprint(f'diam in node {node.ident()} (order {order}) '
                               f'not repaired')
                        err += 1
                else:
                    r = np.array([morph.radii(sec).mean()
                                  for sec in morph.root.sections()
                                  if sec[0].type() == point_type
                                  and sec[0].order() == order]).mean()
                    node.v[SWC.R] = r
        elif args.diam_mode == 'breadth':
            for node in nodes:
                point_type = node.type()
                breadth = node.breadth()
                if args.pool:
                    radii = [m.radii(sec).mean() for m in pool
                             for sec in m.root.sections()
                             if sec[0].type() in types
                             and sec[0].breadth() == breadth]
                    if radii:
                        r = np.mean(radii)
                        node.v[SWC.R] = r
                    else:
                        vprint(f'diam in node {node.ident()} '
                               f'(breadth {breadth}) not repaired')
                        err += 1
                else:
                    r = np.array([morph.radii(sec).mean()
                                  for sec in morph.root.sections()
                                  if sec[0].type() == point_type
                                  and sec[0].breadth() == breadth]).mean()
                    node.v[SWC.R] = r

    if args.seed:
        np.random.seed(args.seed)

    if args.cut:  # pylint: disable=too-many-nested-blocks
        types = {x.type() for x in morph.root.walk() if x.ident() in args.cut}
        for point_type in types:
            intact_branches = dict()
            if args.pool:
                for rec in pool:
                    sections = filter(lambda x: x[0].type() == point_type,
                                      rec.root.sections())
                    nodes = chain(x[0] for x in sections)
                    for node in nodes:
                        order = node.order()
                        if order not in intact_branches:
                            intact_branches[order] = list()
                        intact_branches[order].append((rec, node))
            else:
                sections = filter(lambda x: x[0].type() == point_type,
                                  morph.root.sections())
                nodes = chain(x[0] for x in sections)

                def is_intact(tree, cuts):
                    leaves = [x.ident() for x in tree.leaves()]
                    return set(leaves).isdisjoint(cuts)

                nodes = filter(lambda x: is_intact(x, args.cut), nodes)
                for node in nodes:
                    order = node.order()
                    if order not in intact_branches:
                        intact_branches[order] = list()
                    intact_branches[order].append((morph, node))

            nodes = [x for x in morph.root.walk() if x.type() == point_type
                     and x.ident() in args.cut]
            for node in nodes:
                order = node.order()
                vprint(f'repairing node {node.ident()} (order {order})',
                       end=' ')
                if order in intact_branches:
                    idx = np.random.choice(len(intact_branches[order]))
                    rec, rep = intact_branches[order][idx]
                    vprint(f'using {rep.ident()} (order {order}) ...', end=' ')
                    done = repair_branch(morph, node, rec, rep,
                                         force=args.force_repair)
                    err += 1 if not done else 0
                    vprint('done') if done else vprint('not repaired')
                elif order - 1 in intact_branches:
                    idx = np.random.choice(len(intact_branches[order-1]))
                    rec, rep = intact_branches[order-1][idx]
                    vprint(f'using {rep.ident()} (order {order-1}) ...',
                           end=' ')
                    done = repair_branch(morph, node, rec, rep,
                                         force=args.force_repair)
                    err += 1 if not done else 0
                    vprint('done') if done else vprint('not repaired')
                elif args.force_repair:
                    if intact_branches:
                        order = np.random.choice(list(intact_branches.keys()))
                        idx = np.random.choice(len(intact_branches[order]))
                        rec, rep = intact_branches[order][idx]
                        vprint(f'using {rep.ident()} (order {order}) ...',
                               end=' ')
                        done = repair_branch(morph, node, rec, rep, force=True)
                        err += 1 if not done else 0
                        vprint('done') if done else vprint('not repaired')
                    else:
                        err += 1
                        vprint('... no intact branches, not repaired')
                else:
                    err += 1
                    vprint('... not repaired')

    if args.delete and not args.cut:
        nodes = [x for x in morph.root.walk() if x.ident() in args.delete]
        for node in nodes:
            morph.delete(node)

    if args.delete or args.cut:
        morph = Morph(data=morph.data)

    if args.res:
        ident = 1
        data = list()
        idmap = {-1: -1}
        for sec in filter(lambda x: x[0].type() == SWC.SOMA,
                          morph.root.sections()):
            for node in sec:
                v = node.v.copy()
                i, p = v[SWC.I].astype(int), v[SWC.P].astype(int)
                v[SWC.I], v[SWC.P] = ident, idmap[p]
                idmap[i] = ident
                data.append(v)
                ident += 1
        for sec in filter(lambda x: x[0].type() in
                          set(SWC.TYPES).difference((SWC.SOMA,)),
                          morph.root.sections()):
            length = morph.length(sec)
            points = morph.points(sec)
            # pylint: disable=unsubscriptable-object
            points = np.insert(points, 0, sec[0].parent.v[SWC.XYZR], axis=0)
            points = sample(points, np.rint(length/args.res).astype(int))
            points = points[1:]
            start = True
            for ident, point in enumerate(points, ident):
                x, y, z, r = point
                pid = idmap[sec[0].parent_ident()] if start else ident - 1
                v = np.array([ident, sec[0].type(), x, y, z, r, pid])
                start = False if start else start
                data.append(v)
            idmap[sec[-1].v[SWC.I]] = ident
            ident += 1
        morph = Morph(data=data)

    morph.save(args.out)
    return err



"""Implementation of CLI find command."""

from treem.morph import Morph
from treem.io import SWC


def find(args):  # pylint: disable=too-many-branches
    """Locates single nodes in morphology reconstruction."""
    morph = Morph(args.file)
    types = args.type if args.type else SWC.TYPES
    nodes = filter(lambda x: x.type() in types, morph.root.walk())
    if args.order:
        nodes = filter(lambda x: x.order() in args.order, nodes)
    if args.breadth:
        nodes = filter(lambda x: x.breadth() in args.breadth, nodes)
    if args.degree:
        nodes = filter(lambda x: x.degree() in args.degree, nodes)
    if args.stem:
        nodes = filter(lambda x: x.is_stem() and x.type() != SWC.SOMA, nodes)
    if args.diam is not None:
        if args.compare == 'gt':
            nodes = filter(lambda x: x.diam() > args.diam, nodes)
        elif args.compare == 'lt':
            nodes = filter(lambda x: x.diam() < args.diam, nodes)
        elif args.compare == 'eq':
            nodes = filter(lambda x: x.diam() == args.diam, nodes)
    if args.length is not None:
        if args.compare == 'gt':
            nodes = filter(lambda x: x.length() > args.length, nodes)
        elif args.compare == 'lt':
            nodes = filter(lambda x: x.length() < args.length, nodes)
        elif args.compare == 'eq':
            nodes = filter(lambda x: x.length() == args.length, nodes)
    if args.jump is not None:
        if args.compare == 'gt':
            nodes = filter(lambda x: not x.is_root()
                           and abs((x.parent.coord()-x.coord())[2])
                           > args.jump, nodes)
        elif args.compare == 'lt':
            nodes = filter(lambda x: not x.is_root()
                           and abs((x.parent.coord()-x.coord())[2])
                           < args.jump, nodes)
        elif args.compare == 'eq':
            nodes = filter(lambda x: not x.is_root()
                           and abs((x.parent.coord()-x.coord())[2])
                           == args.jump, nodes)
    if args.cut:
        nodes = filter(lambda x: x.is_leaf(), nodes)
        if not args.bottom_up:
            zcut = max(x.coord()[2] for x in morph.root.walk())
            nodes = filter(lambda x: x.coord()[2] > zcut - args.cut, nodes)
        else:
            zcut = min(x.coord()[2] for x in morph.root.walk())
            nodes = filter(lambda x: x.coord()[2] < zcut + args.cut, nodes)

    if args.sec:
        nodes = filter(lambda x: x.parent.is_fork()
                       or x.parent.is_root(), nodes)

    for node in nodes:
        print(node.ident(), end=' ')
    print()



"""SWC data format defintion and services."""

import json

import numpy as np


class SWC():  # pylint: disable=too-few-public-methods
    """Definitions of the data format."""
    TYPES = (SOMA, AXON, DEND, APIC) = range(1, 5)
    COLS = (I, T, X, Y, Z, R, P) = range(7)
    XYZ = slice(2, 5)
    XYZR = slice(2, 6)
    RADII = slice(5, 6)


class TreemEncoder(json.JSONEncoder):
    """Extended JSONEncoder to serialize treem objects."""
    def default(self, obj):  # pylint: disable=arguments-differ
        if hasattr(obj, 'tolist'): 
            return obj.tolist()
        elif isinstance(obj, set):
            return list(obj)
        else:
            return json.JSONEncoder.default(self, obj)


def load_swc(source):
    """Reads data from SWC file."""
    return np.loadtxt(source)


def save_swc(target, data):
    """Writes data to SWC file."""
    fmt = '%d %d %g %g %g %g %d'
    return np.savetxt(target, data, fmt=fmt)
